/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, Hongbo Zhu
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of Hongbo Zhu
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TDecBinPACoder.cpp
    \brief    binary entropy decoder of PA-Coder
*/

#include "TDecBinPACoder.h"


TDecBinPACoder::TDecBinPACoder()
: m_pcTComBitstream( 0 )
{
}


TDecBinPACoder::~TDecBinPACoder()
{
}


Void
TDecBinPACoder::init( TComBitstream* pcTComBitstream )
{
  m_pcTComBitstream = pcTComBitstream;
}


Void
TDecBinPACoder::uninit()
{
  m_pcTComBitstream = 0;
}


Void
TDecBinPACoder::start()
{
  m_pcTComBitstream->setModeSbac();

	logprob = 0;
  value   = 0;
	xReadByte(value);
	xReadByte(value);
	xReadByte(value);
	xReadByte(value);
}


Void
TDecBinPACoder::finish()
{
}


Void
TDecBinPACoder::decodeBin( UInt& ruiBin, ContextModel &rcCtxModel )
{
	UInt logprobMPS = TComPACoderTables::lgprobMLps[rcCtxModel.getState()][1];
	UInt thrshld = dvsn_interval(logprobMPS+logprob);

	if(thrshld <= value)
	{
		value -= thrshld;
		logprob += TComPACoderTables::lgprobMLps[rcCtxModel.getState()][0];
		ruiBin      = 1 - rcCtxModel.getMps();
		rcCtxModel.updateLPS();
	}
	else
	{
		logprob += logprobMPS;
		ruiBin      = rcCtxModel.getMps();
		rcCtxModel.updateMPS();
	}

	if(logprob >= 0xc00)
	{
		logprob -= 0xc00;
		xReadByte(value);
	  xReadByte(value);
		xReadByte(value);
		renorm_logprob();
	}
}


Void
TDecBinPACoder::decodeBinEP( UInt& ruiBin )
{
	logprob += 128;
	UInt thrshld = dvsn_interval(logprob);

	ruiBin = 0;
	if(thrshld <= value)
	{
		value -= thrshld;
		ruiBin = 1;
	}

	if(logprob >= 0xc00)
	{
		logprob -= 0xc00;
		xReadByte(value);
	  xReadByte(value);
		xReadByte(value);
		renorm_logprob();
	}
}


Void
TDecBinPACoder::decodeBinTrm( UInt& ruiBin )
{
	logprob += 1;
	UInt thrshld = dvsn_interval(logprob);

  if(thrshld <= value)
  {
    ruiBin = 1;
  }
  else
  {
    ruiBin = 0;
		if(logprob >= 0xc00)
		{
			logprob -= 0xc00;
			xReadByte(value);
			xReadByte(value);
			xReadByte(value);
			renorm_logprob();
		}
  }
}


Void  
TDecBinPACoder::xReadByte( UInt& ruiVal )
{
  UInt uiBit = 0;
  m_pcTComBitstream->read(8, uiBit );
  ruiVal  = ( ruiVal << 8 ) | uiBit;
}

Void TDecBinPACoder::renorm_logprob()
{
	Int logprb = logprob;
	Int sel = logprb&127, shft = logprb>>7, rsft = 7-shft;

	UInt val22 = TComPACoderTables::subinterval_division[sel]>>shft, valact = val22&0xff00, val33;
	Int diff2;
	val22 &= 0xff;
	diff2 = ((val22+(1<<rsft))>>(1+rsft))+1;

	logprb += diff2;
	val33 = TComPACoderTables::subinterval_division[logprb&127]>>(logprb>>7);
	if(val33 > valact)
		diff2 += ((diff2+shft+1)>>2);
	else if(sel<30)
		diff2 -= (diff2>>3);

	logprob += diff2;
}





